#include "cloog_util.h"

#define ALLOC(type) (type*)malloc(sizeof(type))

char ** cloog_util_generate_names(int nb_items, char * prefix)
{
    int i;
    char ** names;
    
    names = (char **)malloc(nb_items*sizeof(char *));
    if (names == NULL)  {
	fprintf(stderr, "ERROR: memory overflow.\n");
	exit(1) ;
    }
    for (i=0;i<nb_items;i++) {
	names[i] = (char *)malloc(MAX_NAME*sizeof(char));
	if (names[i] == NULL) {
	    fprintf(stderr, "ERROR: memory overflow.\n");
	  exit(1);
	}
	if (prefix == NULL)
	    sprintf(names[i],"%d",i);
	else
	    sprintf(names[i],"%s%d",prefix,i);
    }
  
    return names ;
}

char **cloog_util_parameter_names(const seq<pdg::parameter>& params)
{
    int nparam = params.size();
    char **names;

    names = (char**)malloc(nparam * sizeof(char*));
    assert(names);
    for (int i = 0; i < nparam; ++i)
	names[i] = strdup(params[i]->name->s.c_str());

    return names;
}

clast_stmt *cloog_util_create_clast(CloogProgram *p, CloogOptions *options)
{
    return cloog_clast_create(p, options);
}

static CloogDomain *cloog_util_domain_empty(int dim)
{
    ::Matrix * matrix = cloog_matrix_alloc(0, 2 + dim);
    CloogDomain * domain = cloog_domain_matrix2domain(matrix);
    cloog_matrix_free(matrix);

    return domain;
}

void cloog_program_builder::init()
{
    p = cloog_program_malloc();
    assert(p != NULL);
    p->names = cloog_names_malloc();
    assert(p->names != NULL);
    p->language = 'c';

    next = &p->loop;
    nextBL = &p->blocklist;

    scatterings = NULL;
    nextScat = &scatterings;
}

cloog_program_builder::cloog_program_builder()
{
    init();
    p->context = cloog_util_domain_empty(0);
    nparam = 0;
    niter = 0;
}

cloog_program_builder::cloog_program_builder(Polyhedron *ctx)
{
    init();
    p->context = cloog_domain_alloc(ctx);
    nparam = ctx->Dimension;
    niter = 0;
}

cloog_program_builder::cloog_program_builder(pdg::PDG *pdg)
{
    init();
    nparam = pdg->params.size();
    niter = 0;
    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);
    p->context = cloog_domain_alloc(ctx);

    int i;
    for (i = 0, niter = 0; i < pdg->nodes.size(); ++i) {
	if (!pdg->nodes[i]->scattering) {
	    fprintf(stderr, "Missing scattering function\n");
	    exit(-1);
	}

	add_domain(*pdg->nodes[i]->source, pdg->nodes[i]->statement->operation,
		   pdg->nodes[i]);
	/*
	value_init(l->stride);
	value_set_si(l->stride, 1);
	*/

	add_scattering(*pdg->nodes[i]->scattering);
    }

    int nscat = cloog_domain_dimension(scatterings->domain) - 
			cloog_domain_dimension(p->loop->domain);
    set_names(pdg->params, nscat);

    cloog_program_scatter(p, scatterings);
}

void cloog_program_builder::set_names(const seq<pdg::parameter>& params, int nscat)
{
    int nparam = params.size();

    p->names->nb_parameters = nparam;
    p->names->parameters = cloog_util_parameter_names(params);

    p->names->nb_iterators = niter;
    p->names->iterators = cloog_util_generate_names(niter, "i");

    p->nb_scattdims = nscat;
    if (nscat) {
	p->scaldims = (int *)malloc(p->nb_scattdims*(sizeof(int)));
	for (int i = 0; i < p->nb_scattdims; i++)
	    p->scaldims[i] = 0;
	p->names->scattering = cloog_util_generate_names(p->nb_scattdims, "c");
    } else {
	p->scaldims = NULL;
	p->names->scattering = NULL;
    }

    p->names->nb_scattering = p->nb_scattdims;
    p->names->nb_scalars = 0;
}

void cloog_program_builder::add_domain(Polyhedron *P, int nr, void *data)
{
    CloogStatement * statement;
    CloogLoop *l = cloog_loop_malloc();
    l->next = NULL;
    l->inner = NULL;
    l->domain = cloog_domain_alloc(P);

    int n = cloog_domain_dimension(l->domain) - nparam;
    if (n > niter)
	niter = n;

    statement = cloog_statement_alloc(nr);
    statement->usr = data;
    int nb_iterators = cloog_domain_dimension(l->domain) - p->names->nb_parameters;
    l->block = cloog_block_alloc(statement, NULL, 0, NULL, nb_iterators);

    *next = l;
    *nextBL = cloog_block_list_alloc((*next)->block);
    next = &(*next)->next;
    nextBL = &(*nextBL)->next;
}

void cloog_program_builder::add_scattering(Polyhedron *S)
{
    CloogDomainList *s = ALLOC(CloogDomainList);
    s->domain = cloog_domain_alloc(S);
    s->next = NULL;

    *nextScat = s;
    nextScat = &(s->next);
}

cloog_program_builder::~cloog_program_builder()
{
    if (scatterings)
	cloog_domain_list_free(scatterings);
}

static clast_expr *clast_expr_copy(clast_expr *e)
{
    if (!e)
	return NULL;
    switch (e->type) {
    case clast_expr::expr_term: {
	clast_term* t = (struct clast_term*) e;
	return &new_clast_term(t->val, t->var)->expr;
    }
    case clast_expr::expr_red: {
	clast_reduction *r = (struct clast_reduction*) e;
	clast_reduction *r2 = new_clast_reduction(r->type, r->n);
	for (int i = 0; i < r->n; ++i)
	    r2->elts[i] = clast_expr_copy(r->elts[i]);
	return &r2->expr;
    }
    case clast_expr::expr_bin: {
	clast_binary *b = (struct clast_binary*) e;
	return &new_clast_binary(b->type, clast_expr_copy(b->LHS), b->RHS)->expr;
    }
    default:
	assert(0);
    }
}

static void simple_guards(clast_for *s);
static void simple_guards(clast_guard *s);

void cloog_util_simple_guards(clast_stmt *s)
{
    for ( ; s; s = s->next) {
	switch (s->type) {
	case clast_stmt::stmt_root:
	case clast_stmt::stmt_ass:
	case clast_stmt::stmt_user:
	    break;
	case clast_stmt::stmt_for:
	    simple_guards((clast_for *) s);
	    break;
	case clast_stmt::stmt_guard:
	    simple_guards((clast_guard *) s);
	    break;
	default:
	    assert(0);
	}
    }
}

void convert_minmax_to_ifs(clast_expr **rp, clast_expr *LHS, int sign,
			   clast_stmt **body)
{
    assert((*rp)->type == clast_expr::expr_red);
    clast_reduction *r = (clast_reduction *)*rp;

    clast_stmt *child = *body;
    while (r->n > 1) {
	r->n--;
	clast_guard *newg = new_clast_guard(1);
	newg->eq[0].LHS = clast_expr_copy(LHS);
	newg->eq[0].RHS = r->elts[r->n];
	newg->eq[0].sign = sign;
	newg->then = child;
	child = &newg->stmt;
    }
    *body = child;
    *rp = r->elts[0];
    r->n--;
    free_clast_reduction(r);
}

void simple_guards(clast_for *f)
{
    if (f->LB->type == clast_expr::expr_red) {
	clast_reduction *r = (clast_reduction *)f->LB;
	if (r->type != clast_red_sum) {
	    Value one;
	    value_init(one);
	    value_set_si(one, 1);
	    clast_term *LHS = new_clast_term(one, f->iterator);
	    convert_minmax_to_ifs(&f->LB, &LHS->expr, 1, &f->body);
	    free_clast_term(LHS);
	    value_clear(one);
	}
    }
    if (f->UB && f->UB->type == clast_expr::expr_red) {
	clast_reduction *r = (clast_reduction *)f->UB;
	if (r->type != clast_red_sum) {
	    Value one;
	    value_init(one);
	    value_set_si(one, 1);
	    clast_term *LHS = new_clast_term(one, f->iterator);
	    convert_minmax_to_ifs(&f->UB, &LHS->expr, -1, &f->body);
	    free_clast_term(LHS);
	    value_clear(one);
	}
    }
    cloog_util_simple_guards(f->body);
}

void simple_guards_minmax(clast_guard *g)
{
    if (g->eq[0].RHS->type != clast_expr::expr_red)
	return;

    clast_reduction *r = (clast_reduction *)g->eq[0].RHS;
    if (r->type == clast_red_sum)
	return;

    convert_minmax_to_ifs(&g->eq[0].RHS, g->eq[0].LHS, g->eq[0].sign, &g->then);
}

void simple_guards(clast_guard *g)
{
    clast_stmt *then = g->then;
    clast_stmt *child = g->then;

    while (g->n > 1) {
	g->n--;
	clast_guard *newg = new_clast_guard(1);
	newg->eq[0].LHS = g->eq[g->n].LHS;
	newg->eq[0].RHS = g->eq[g->n].RHS;
	newg->eq[0].sign = g->eq[g->n].sign;
	newg->then = child;
	simple_guards_minmax(newg);
	child = &newg->stmt;
    }
    g->then = child;
    simple_guards_minmax(g);

    cloog_util_simple_guards(then);
}

static int clast_equation_equal(clast_equation *e1, clast_equation *e2)
{
    return e1->sign == e2->sign &&
	   clast_expr_equal(e1->LHS, e2->LHS) &&
	   clast_expr_equal(e1->RHS, e2->RHS);
}

struct guard_info {
    struct clast_equation *eq;
    int			   first;
    int			   last;
};

/* attach tail to the last element of head and return this previously
 * last element of head.
 */
static clast_stmt *append_stmt_list(clast_stmt *head, clast_stmt *tail)
{
    for ( ; head; head = head->next)
	if (!head->next) {
	    head->next = tail;
	    return head;
	}
}

/* Hoist info->eq from the info->first element of the next chain
 * of s to the info->last element of this chain.
 * The above subchain is replaced by the guard with info->eq as condition
 * and with the subchain as the "then" element.
 * The first element in the subchain is assumed to have a guard among
 * its nested guards with eq[0] equal to info->eq, which is rotated
 * outermost.
 * The other elements are assumed to have a guard with eq[0] equivalent
 * to info->eq and these are simply removed.
 */
static clast_stmt *hoist(struct guard_info* info, clast_stmt *s)
{
    clast_stmt **parent = &s;
    clast_stmt *sf = s, *s2, *next;
    clast_guard *hoist, *up;

    for (int i = 0; i < info->first; ++i) {
	parent = &sf->next;
	sf = sf->next;
    }

    /* Find guard and rotate out */
    assert(sf->type == clast_stmt::stmt_guard);
    up = hoist = (clast_guard *)sf;
    while (&hoist->eq[0] != info->eq) {
	clast_stmt *s2 = hoist->then;
	assert(!s2->next);
	assert(s2->type == clast_stmt::stmt_guard);
	up = hoist;
	hoist = (clast_guard *)s2;
    }
    if (up != hoist) {
	*parent = &hoist->stmt;
	up->then = hoist->then;
	hoist->then = sf;
	hoist->stmt.next = sf->next;
	sf->next = NULL;
    }

    /* Move other elements in the chain to the then part of the hoisted if
     * and remove the equivalent guard.
     */
    s2 = &hoist->stmt;
    s2 = append_stmt_list(hoist->then, s2->next);
    for (int i = info->first+1; i <= info->last; ++i) {
	clast_stmt **parent = &s2->next;
	clast_guard *current;
	s2 = s2->next;
	assert(s2->type == clast_stmt::stmt_guard);
	current = (clast_guard *)s2;
	while (!clast_equation_equal(&current->eq[0], info->eq)) {
	    parent = &current->then;
	    clast_stmt *s3 = current->then;
	    assert(!s3->next);
	    assert(s3->type == clast_stmt::stmt_guard);
	    current = (clast_guard *)s3;
	}
	if (*parent == s2)
	    s2 = append_stmt_list(current->then, s2->next);
	else
	    assert(!current->stmt.next);
	*parent = current->then;
	current->then = NULL;
	free_clast_guard(current);
    }
    hoist->stmt.next = s2->next;
    s2->next = NULL;

    return s;
}

/* Hoist ifs that are common to several adjacent statements.
 * 
 * We first create a list of guards and note where they first
 * and last appear in sequence.  Then we hoist the if with largest
 * sequence and repeat until the largest sequence is of length 1.
 */
static clast_stmt *do_hoist_ifs(clast_stmt *s)
{
    int n = 0, m = 0;
    for (clast_stmt *s2 = s; s2; s2 = s2->next) {
	clast_stmt *g = s2;
	int has_guard = 0;
	while (g->type == clast_stmt::stmt_guard) {
	    has_guard = 1;
	    n++;
	    g = ((clast_guard *) g)->then;
	    if (g->next)
		break;
	}
	if (has_guard)
	    ++m;
    }
    if (m <= 1)
	return s;

    struct guard_info info[n];
    int i = 0, j = 0, max;
    for (clast_stmt *s2 = s; s2; s2 = s2->next) {
	clast_stmt *gs = s2;
	while (gs->type == clast_stmt::stmt_guard) {
	    int k;
	    clast_guard *g = (clast_guard *) gs;
	    assert(g->n == 1);
	    for (k = 0; k < i; ++k)
		if (info[k].last == j-1 && 
			clast_equation_equal(info[k].eq, &g->eq[0]))
		    break;
	    if (k == i) {
		info[k].eq = &g->eq[0];
		info[k].first = j;
		i++;
	    }
	    info[k].last = j;
	    gs = g->then;
	    if (gs->next)
		break;
	}
	++j;
    }
    n = i;
    do {
	max = 0;
	for (i = 1; i < n; ++i)
	    if (info[i].last - info[i].first > 
		    info[max].last - info[max].first)
		max = i;
	if (info[max].last == info[max].first)
	    break;
	int first = info[max].first;
	int last = info[max].last;
	s = hoist(&info[max], s);
	for (int i = 0; i < n; ++i) {
	    if (info[i].last < first)
		continue;
	    if (info[i].first > last) {
		info[i].first -= last-first;
		info[i].last -= last-first;
		continue;
	    }
	    info[i--] = info[--n];
	}
    } while (n > 0);

    return s;
}

clast_stmt *cloog_util_hoist_ifs(clast_stmt *s)
{
    if (s->next)
	s = do_hoist_ifs(s);

    for (clast_stmt *s2 = s; s2; s2 = s2->next) {
	switch (s2->type) {
	case clast_stmt::stmt_root:
	case clast_stmt::stmt_ass:
	case clast_stmt::stmt_user:
	    break;
	case clast_stmt::stmt_for: {
	    clast_for *f = (clast_for *) s2;
	    f->body = cloog_util_hoist_ifs(f->body);
	    break;
	}
	case clast_stmt::stmt_guard: {
	    clast_guard *g = (clast_guard *) s2;
	    g->then = cloog_util_hoist_ifs(g->then);
	    break;
	}
	default:
	    assert(0);
	}
    }
    return s;
}
