#include <vector>
#include <iostream>

#include "yaml.h"
#include "pdg.h"
#include "da.h"

extern "C" {
#include "piputil.h"
}

using pdg::PDG;
using namespace std;

namespace da {

struct na_triple {
    pdg::node 	*node;
    pdg::access	*access;
    int		 parts;
    na_triple(pdg::node *n, pdg::access *a, int p) : node(n), access(a), parts(p) {}
};

static int precedes_level_nodes(na_triple *first, na_triple *second);
static int precedes_level_accesses(na_triple *first, na_triple *second);

static Polyhedron *convert_access(std::vector< std::vector<int> > &map,
				  std::vector< std::vector<int> > &dom,
				  int in, int out, int nparam)
{
    assert(map.size() >= 1);
    int extra_map = map[0].size()-2 - in - out - nparam;
    int extra_dom = dom.size() ? dom[0].size()-2 - in - nparam : 0;
    Matrix *M = Matrix_Alloc(map.size() + dom.size(),
			     2 + in + out + extra_map + extra_dom + nparam);
    for (int i = 0; i < map.size(); ++i) {
	int j;
	for (j = 0; j < 1 + in + out + extra_map; ++j)
	    value_set_si(M->p[i][j], map[i][j]);
	for (; j < map[0].size(); ++j)
	    value_set_si(M->p[i][j+extra_dom], map[i][j]);
    }
    for (int i = 0; i < dom.size(); ++i) {
	int j;
	for (j = 0; j < 1 + in; ++j)
	    value_set_si(M->p[map.size()+i][j], dom[i][j]);
	for (; j < dom[0].size(); ++j)
	    value_set_si(M->p[map.size()+i][j+out+extra_map], 
			 dom[i][j]);
    }
    Polyhedron *R = Constraints2Polyhedron(M, POL_NO_DUAL);
    Matrix_Free(M);
    return R;
}

/* Combine constraints of the "pure" mapping with the constraints
 * on the domain.
 */
static int convert_access(pip_rel *pip_a, na_triple *na, int nparam)
{
    int n = 0;
    for (int i = 0; i < na->access->map->constraints.size(); ++i)
	for (int j = 0; j < na->node->source->constraints.size(); ++j) {
	    Polyhedron *R;
	    std::vector< std::vector<int> > &map = 
		    na->access->map->constraints[i]->el;
	    std::vector< std::vector<int> > &dom = 
		    na->node->source->constraints[j]->el;

	    pip_a[n].n_input = na->access->map->input;
	    pip_a[n].n_output = na->access->map->output;
	    R = convert_access(map, dom, pip_a[n].n_input, pip_a[n].n_output, nparam);
	    pip_a[n].data = na;
	    pip_a[n].rel = R;
	    ++n;
	}
    return n;
}

static pdg::UnionMap *reattach_dependence(pdg::dependence *d)
{
    pdg::UnionMap *moved = new pdg::UnionMap;
    pdg::UnionMap *rel = d->relation;
    moved->input = moved->output = rel->output;
    moved->params = rel->params;
    for (int i = 0; i < rel->constraints.size(); ++i) {
	pdg::Matrix *oldM = rel->constraints[i];
	pdg::Matrix *M = new pdg::Matrix;
	assert(oldM->el.size() > 0);
	int cols = oldM->el[0].size() - rel->input + moved->input;
	M->el.insert(M->el.begin(), oldM->el.size() - rel->input + moved->input,
		     std::vector<int>(cols));
	for (int j = 0; j < moved->input; ++j) {
	    M->el[j][1+j] = 1;
	    M->el[j][1+moved->input+j] = -1;
	}
	for (int j = 0; j < oldM->el.size() - rel->input; ++j) {
	    M->el[moved->input+j][0] = oldM->el[rel->input+j][0];
	    for (int k = 1+moved->input; k < cols; ++k)
		M->el[moved->input+j][k] = 
		    oldM->el[rel->input+j][k-moved->input+rel->input];
	}
	moved->constraints.push_back(M);
    }
    return moved;
}

static pdg::UnionMap* copy_unionmap(pdg::UnionMap *rel)
{
    pdg::UnionMap *copy = new pdg::UnionMap;
    copy->input = rel->input;
    copy->output = rel->output;
    copy->params = rel->params;
    for (int i = 0; i < rel->constraints.size(); ++i) {
	pdg::Matrix *oldM = rel->constraints[i];
	pdg::Matrix *M = new pdg::Matrix;
	for (int i = 0; i < oldM->el.size(); ++i)
	    M->el.push_back(oldM->el[i]);
	copy->constraints.push_back(M);
    }
    return copy;
}

static void check_reordering(PDG *pdg, pdg::dependence *dep)
{
    int nparam = pdg->params.size();
    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);
    pdg::UnionMap *rel = dep->relation;
    Polyhedron *P = *rel;
    dep->reordering = pip_is_reordering(P, rel->input, rel->output, ctx);
    Domain_Free(P);
    Polyhedron_Free(ctx);
}

/* Check whether writes that correspond to reads from the same read iteration
 * in odep and ndep are in the same order as the (to) accesses.
 */
static bool is_inorder_accesses(pdg::dependence *odep, pdg::dependence *ndep,
				Polyhedron *ctx)
{
    bool inorder;
    pdg::UnionMap *orel = odep->relation;
    pdg::UnionMap *nrel = ndep->relation;
    Polyhedron *Pn = *nrel;
    Polyhedron *Po = *orel;
    Polyhedron *P1, *P2;
    if (odep->to_access->nr < ndep->to_access->nr) {
	P1 = Po;
	P2 = Pn;
    } else {
	P1 = Pn;
	P2 = Po;
    }
    inorder = pip_is_inorder_accesses(P1, P2, orel->input, orel->output, ctx);
    Domain_Free(Po);
    Domain_Free(Pn);
    return inorder;
}

static bool combined_reordering(PDG *pdg, pdg::dependence *odep, pdg::dependence *ndep)
{
    bool reordering;
    int nparam = pdg->params.size();
    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);
    pdg::UnionMap *orel = odep->relation;
    pdg::UnionMap *nrel = ndep->relation;
    assert(orel->input == nrel->input);
    assert(orel->output == nrel->output);
    Polyhedron *P = Domain_concat(*nrel, *orel);
    reordering = pip_is_reordering(P, orel->input, orel->output, ctx);
    Domain_Free(P);
    if (!reordering) {
	if (odep->type != pdg::dependence::pn_union) {
	    assert(odep->to_access);
	    if (!is_inorder_accesses(odep, ndep, ctx))
		reordering = true;
	} else {
	    for (int j = 0; j < pdg->dependences.size(); ++j) {
		pdg::dependence *part_dep = pdg->dependences[j];
		if (part_dep->type != pdg::dependence::pn_part)
		    continue;
		if (part_dep->container != odep)
		    continue;
		if (!is_inorder_accesses(part_dep, ndep, ctx)) {
		    reordering = true;
		    break;
		}
	    }
	}
    }
    Polyhedron_Free(ctx);
    return reordering;
}

static void combine_pn_channels(PDG *pdg, pdg::dependence *d)
{
    if (d->to == d->from)
	return;

    pdg::dependence *d_union = NULL;
    for (int i = 0; i < d->to->sources.size(); ++i) {
	if (d->to->sources[i]->from_access != d->from_access)
	    continue;
	pdg::dependence *dep = d->to->sources[i];
	if (d->reordering == -1) {
	    check_reordering(pdg, d);
	    if (d->reordering)
		break;
	}
	if (dep->reordering == -1) {
	    check_reordering(pdg, dep);
	    if (dep->reordering) {
		d->to->sources.erase(dep);
		break;
	    }
	}
	if (combined_reordering(pdg, dep, d))
	    continue;
	d_union = d->to->sources[i];
	break;
    }
    if (!d_union) {
	if (d->reordering != 1)
	    d->to->sources.push_back(d);
	return;
    }
    if (d_union->type != pdg::dependence::pn_union) {
	pdg::dependence *copy = new pdg::dependence;
	copy->from = d_union->from;
	copy->to = d_union->to;
	copy->from_access = d_union->from_access;
	copy->to_access = d_union->to_access;
	copy->array = d_union->array;
	copy->type = pdg::dependence::pn_part;
	copy->relation = copy_unionmap(d_union->relation);
	copy->container = d_union;
	d_union->to_access = NULL;
	d_union->type = pdg::dependence::pn_union;
	pdg->dependences.push_back(copy);
    }
    pdg::UnionMap *rel = d->relation;
    assert(rel->input == d_union->relation->input);
    assert(rel->output == d_union->relation->output);
    for (int i = 0; i < rel->constraints.size(); ++i)
	d_union->relation->constraints.push_back(rel->constraints[i]);
    d->relation = copy_unionmap(rel);
    delete rel;
    d->type = pdg::dependence::pn_part;
    d->container = d_union;
}

/* returns true if all input dimensions are equal to the 
 * corresponding output dimensions
 */
static bool is_identity_map(pdg::UnionMap* rel)
{
    assert(rel->input == rel->output);
    int dim = rel->input;
    for (int c = 0; c < rel->constraints.size(); ++c) {
	pdg::Matrix *M = rel->constraints[c];
	int i;
	for (i = 0; i < dim; ++i) {
	    int k, k2;
	    /* look for input[i] == output[i] */
	    for (k = 0; k < M->el.size(); ++k) {
		if (M->el[k][0] != 0)
		    continue;
		if (M->el[k][1+i] != 1)
		    continue;
		if (M->el[k][1+dim+i] != -1)
		    continue;
		int l;
		for (l = 1; l < M->el[k].size(); ++l) {
		    if (l == 1+i)
			continue;
		    if (l == 1+dim+i)
			continue;
		    if (M->el[k][l] != 0)
			break;
		}
		if (l == M->el[k].size())
		    break;
	    }
	    if (k < M->el.size())
		continue;

	    /* look for input[i] == ... and output[i] == ... constraints */
	    for (k = 0; k < M->el.size(); ++k) {
		if (M->el[k][0] != 0)
		    continue;
		if (M->el[k][1+i] == 1)
		    break;
	    }
	    if (k == M->el.size())
		break;
	    for (k2 = 0; k2 < M->el.size(); ++k2) {
		if (M->el[k2][0] != 0)
		    continue;
		if (M->el[k2][1+dim+i] == 1)
		    break;
	    }
	    if (k2 == M->el.size())
		break;
	    int l;
	    for (l = 1; l < M->el[k].size(); ++l) {
		if (l == 1+i)
		    continue;
		if (l == 1+dim+i)
		    continue;
		if (M->el[k][l] != M->el[k2][l])
		    break;
	    }
	    if (l != M->el[k].size())
		break;
	}
	if (i < dim)
	    return false;
    }
    return true;
}

void find_deps(PDG* pdg, pdg::array *a, type t)
{
    int nparam = pdg->params.size();
    enum pdg::dependence::type dtype;
    int selfinput = 0;
    int reuse = 0;
    switch (t) {
    case flow:
	dtype = pdg::dependence::flow;
	break;
    case anti:
	dtype = pdg::dependence::anti;
	break;
    case pn:
    case pn_propagate:
	dtype = pdg::dependence::pn;
	selfinput = 1;
	break;
    case data_reuse:
	dtype = pdg::dependence::reuse;
	reuse = 1;
	break;
    }

    vector<na_triple> readers;
    vector<na_triple> writers;
    int nreaderparts = 0;
    int nwriterparts = 0;
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	pdg::statement *s = pdg->nodes[i]->statement;
	for (int j = 0; j < s->accesses.size(); ++j) {
	    pdg::access *access = s->accesses[j];
	    if (access->array != a)
		continue;
	    int parts = node->source->constraints.size() *
			access->map->constraints.size();
	    switch (t) {
	    case flow:
	    case pn:
	    case pn_propagate:
	    case data_reuse:
	    case anti:
		if ((access->type == pdg::access::read) ^ (t == anti)) {
		    readers.push_back(na_triple(node, access, parts));
		    nreaderparts += parts;
		} else {
		    writers.push_back(na_triple(node, access, parts));
		    nwriterparts += parts;
		}
		break;
	    }
	}
    }

    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);
    pip_dependence_info *dep;
    int maxsize = (selfinput || reuse) ? nwriterparts + nreaderparts : nwriterparts;
    dep = (pip_dependence_info *) malloc(sizeof(pip_dependence_info) + 
				     maxsize * sizeof(pip_rel));
    pip_rel *converted_readers = new pip_rel[nreaderparts];
    dep->ctx = ctx;
    dep->precedes_level = (pip_access_precedes_level)
	    ((t == pn_propagate) ? precedes_level_accesses : precedes_level_nodes);
    for (int i = 0, j = 0; i < writers.size(); ++i) {
	j += convert_access(&dep->write[j], &writers[i], nparam);
    }
    for (int i = 0, j = 0; i < readers.size(); ++i) {
	j +=  convert_access(&converted_readers[j], &readers[i], nparam);
    }
    for (int i = 0; i < nreaderparts; ++i) {
	dep->n_write = nwriterparts;
	dep->read = converted_readers[i];
	if (selfinput && dep->n_write) {
	    pdg::node *readnode = ((na_triple *)converted_readers[i].data)->node;
	    for (int j = 0; j < nreaderparts; ++j) {
		if (((na_triple *)converted_readers[j].data)->node == readnode)
		    dep->write[dep->n_write++] = converted_readers[j];
	    }
	}
	if (reuse) {
	    for (int j = 0; j < nreaderparts; ++j)
		dep->write[dep->n_write++] = converted_readers[j];
	}
	pip_rel* pr = pip_lastwrite(dep);
	for (int j = 0; j < dep->n_write; ++j) {
	    if (!pr[j].rel)
		continue;
	    pdg::dependence *d = new pdg::dependence;
	    d->array = a;
	    d->type = dtype;
	    d->from = ((na_triple*)dep->write[j].data)->node;
	    d->to = ((na_triple *)converted_readers[i].data)->node;
	    d->from_access = ((na_triple*)dep->write[j].data)->access;
	    d->to_access = ((na_triple *)converted_readers[i].data)->access;
	    d->relation = new pdg::UnionMap(pr[j].n_input, pr[j].n_output,
					    pr[j].rel, pdg->params);
	    Domain_Free(pr[j].rel);
	    pdg->dependences.push_back(d);
	    if (t == pn_propagate) {
		if (d->from == d->to && is_identity_map(d->relation))
		    d->type = pdg::dependence::pn_wire;
		else
		    combine_pn_channels(pdg, d);
	    }
	}
	if (pr[dep->n_write].rel) {
	    pdg::dependence *d = new pdg::dependence;
	    d->array = a;
	    d->type = pdg::dependence::uninitialized;
	    d->to = ((na_triple *)converted_readers[i].data)->node;
	    d->to_access = ((na_triple *)converted_readers[i].data)->access;
	    d->relation = new pdg::UnionMap(0, pr[dep->n_write].n_output,
					      pr[dep->n_write].rel, pdg->params);
	    pdg->dependences.push_back(d);
	}
	if (pr[dep->n_write].rel)
	    Domain_Free(pr[dep->n_write].rel);
	free(pr);
    }
    for (int i = 0; i < nwriterparts; ++i) {
	Domain_Free(dep->write[i].rel);
    }
    for (int i = 0; i < nreaderparts; ++i) {
	Domain_Free(converted_readers[i].rel);
    }
    delete [] converted_readers;
    free(dep);
    Polyhedron_Free(ctx);
}

static int precedes_level_nodes(na_triple *first, na_triple *second)
{
    int d = 0;
    int cmp = 0;
    for (int i = 0; i < first->node->prefix.size(); ++i) {
	if (i >= second->node->prefix.size())
	    break;
	cmp = first->node->prefix[i] - second->node->prefix[i];
	if (cmp)
	    break;
	if (first->node->prefix[i] == -1)
	    ++d;
    }
    return 2*d-1+(cmp<0);
}

static int precedes_level_accesses(na_triple *first, na_triple *second)
{
    int cmp = 0;
    int d = 0;
    for (int i = 0; i < first->node->prefix.size(); ++i) {
	if (i >= second->node->prefix.size())
	    break;
	cmp = first->node->prefix[i] - second->node->prefix[i];
	if (cmp)
	    break;
	if (first->node->prefix[i] == -1)
	    ++d;
    }
    /* same node; now compare accesses */
    if (!cmp)
	cmp = first->access->nr - second->access->nr;
    return 2*d-1+(cmp<0);
}

}
