#include <vector>
#include <bernstein/bernstein.h>
#include <bernstein/piecewise_lst.h>
#include <barvinok/bernstein.h>
#include <ginac/ginac.h>
#include "yaml.h"
#include "pdg.h"
#include "da.h"
#include "evalue_util.h"
#include "translation.h"
extern "C" {
#include "piputil.h"
}

#define MAXRAYS    POL_NO_DUAL

using std::vector;
using pdg::PDG;
using namespace da;
using namespace GiNaC;
using namespace barvinok;
using namespace bernstein;
using std::cout;
using std::cerr;
using std::endl;

static evalue *compute_uninitialized_read(evalue *total, Polyhedron *domain,
					  Polyhedron *reads, int dim,
					  Polyhedron *ctx)
{
    evalue mone;
    value_init(mone.d);
    evalue_set_si(&mone, -1, 1);

    for (Polyhedron *P = domain; P; P = P->next) {
	evalue *EC = lexsmaller_domain(reads, P, dim, ctx);
	emul(&mone, EC);

	if (EVALUE_IS_ZERO(*EC)) {
	    free_evalue_refs(EC);
	    free(EC);
	    continue;
	}

	if (!total)
	    total = EC;
	else {
	    eadd(EC, total);
	    reduce_evalue(total);
	    free_evalue_refs(EC);
	    free(EC);
	}
    }

    free_evalue_refs(&mone);
    return total;
}

static evalue *compute_live(evalue *total, Polyhedron *domain,
			    Polyhedron *dep, int dim, Polyhedron *ctx)
{
    evalue mone;
    value_init(mone.d);
    evalue_set_si(&mone, -1, 1);

    Polyhedron *reads = NULL;
    for (Polyhedron *P = dep; P; P = P->next) {
	Polyhedron *read = Polyhedron_projectout_dep(P, dim);
	reads = Domain_concat(read, reads);
    }

    for (Polyhedron *P = domain; P; P = P->next) {
	evalue *EP = lexsmaller_domain(dep, P, dim, ctx);
	evalue *EC = lexsmaller_domain(reads, P, dim, ctx);
	emul(&mone, EC);
	eadd(EC, EP);
	free_evalue_refs(EC);
	free(EC);
	reduce_evalue(EP);

	if (EVALUE_IS_ZERO(*EP)) {
	    free_evalue_refs(EP);
	    free(EP);
	    continue;
	}

	if (!total)
	    total = EP;
	else {
	    eadd(EP, total);
	    reduce_evalue(total);
	    free_evalue_refs(EP);
	    free(EP);
	}
    }

    Domain_Free(reads);
    free_evalue_refs(&mone);
    return total;
}

pdg::UnionSet *scatter_domain(pdg::node *node)
{
    pdg::UnionMap *s = node->scattering;
    pdg::UnionSet *scattered = new pdg::UnionSet;
    scattered->dim = s->input;
    assert(s->constraints.size() == 1);
    pdg::Matrix *SM = s->constraints[0];
    assert(SM->el.size());
    int nparam = node->source->params.size();
    for (int i = 0; i < node->source->constraints.size(); ++i) {
	pdg::Matrix *NM = node->source->constraints[i];
	pdg::Matrix *M = new pdg::Matrix;
	M->el = NM->el;
	for (int j = 0; j < M->el.size(); ++j)
	    M->el[j].insert(M->el[j].begin()+1, s->input, 0);
	int cols = M->el.size() ? M->el[0].size() : SM->el[0].size();
	M->el.insert(M->el.begin(), s->input, vector<int>(cols));

	int scols = SM->el[0].size();
	for (int j = 0; j < s->input; ++j) {
	    assert(SM->el[j][1+j] == 1);
	    M->el[j][1+j] = 1;
	    for (int k = 0; k < s->output; ++k)
		M->el[j][1+s->input+k] = SM->el[j][1+s->input+k];
	    for (int k = 0; k < nparam+1; ++k)
		M->el[j][cols-(nparam+1)+k] = SM->el[j][scols-(nparam+1)+k];
	}

	scattered->constraints.push_back(M);
    }
    return scattered;
}

static piecewise_lst *maximize(evalue *E, Polyhedron *ctx, const exvector& params)
{
    piecewise_lst *pl = evalue_bernstein_coefficients(NULL, E, ctx, params);

    pl->maximize();
    pl->simplify_domains(ctx, MAXRAYS);

    return pl;
}

static exvector construct_params(pdg::PDG *pdg)
{
    int nparam = pdg->params.size();
    const char *param_names[nparam];
    for (int i = 0; i < nparam; ++i)
	param_names[i] = pdg->params[i]->name->s.c_str();
    return constructParameterVector(param_names, nparam);
}

int main(int argc, char * argv[])
{
    PDG *pdg;

    pdg = yaml::Load<PDG>(stdin);
    int nparam = pdg->params.size();

    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);

    for (int i = 0; i < pdg->arrays.size(); ++i)
	find_deps(pdg, pdg->arrays[i], data_reuse);

    set_scattering_from_prefix(pdg);
    Polyhedron *domain = NULL;
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	pdg::UnionSet *s = scatter_domain(node);
	Polyhedron *P = *s;
	Polyhedron *D;
	D = Polyhedron_projectout(P, s->dim,
				  P->Dimension - s->dim - nparam, nparam);
	Polyhedron_Free(P);
	domain = Domain_concat(D, domain);
	s->free();
	delete s;
    }

    exvector params = construct_params(pdg);

    evalue *inputsize = evalue_zero();
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type != pdg::dependence::uninitialized)
	    continue;
	pdg::UnionMap *rel = dep->relation;
	Polyhedron *D = *rel;
	for (Polyhedron *P = D; P; P = P->next) {
	    evalue *EP = barvinok_enumerate_e(P, 
			    P->Dimension - rel->output - nparam, nparam, MAXRAYS);
	    eadd(EP, inputsize);
	    free_evalue_refs(EP);
	    free(EP);
	}
	Domain_Free(D);
    }
    evalue_simplify_domains(inputsize, ctx);
    ex ex_inputsize = 0;
    if (!EVALUE_IS_ZERO(*inputsize)) {
	assert(value_zero_p(inputsize->d));
	assert(inputsize->x.p->size == 2);
	assert(universeQ(EVALUE_DOMAIN(inputsize->x.p->arr[0])));
	ex_inputsize = evalue2ex(&inputsize->x.p->arr[1], params);
    }
    free_evalue_refs(inputsize);
    free(inputsize);

    evalue *total = NULL;
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	pdg::UnionMap *rel = scatter_dependence(dep);
	Polyhedron *P = *rel;
	if (dep->type == pdg::dependence::uninitialized) {
	    total = compute_uninitialized_read(total, domain, P, rel->output, ctx);
	} else {
	    assert(rel->input == rel->output);
	    total = compute_live(total, domain, P, rel->input, ctx);
	}
	Domain_Free(P);
	rel->free();
	delete rel;
    }
    piecewise_lst *pl = maximize(total, ctx, params);
    free_evalue_refs(total);
    free(total);
    pl->add(ex_inputsize);
    cout << *pl << endl;
    delete pl;

    Domain_Free(domain);
    Polyhedron_Free(ctx);

    pdg->free();
    delete pdg;

    return 0;
}
