#include "pdgutil.h"

namespace pdg {

std::vector<Global_Var_Decl *> extract_globals(PDG* pdg)
{
    std::vector<Global_Var_Decl *> globals;
    for (int i = 0; i < pdg->params.size(); ++i)
	globals.push_back(new Global_Var_Decl(pdg->params[i]->name->s.c_str()));
    return globals;
}

Relation constraints2relation(
    int in, int out, const std::vector<Global_Var_Decl *>& params,
    std::vector< std::vector<int> > & c)
{
    Relation r (in, out);
    assert(c.size() >= 1);
    int	dim = c[0].size();
    varvector vars;
    int nparam = params.size();

    F_Exists *e = r.add_exists();
    F_And *base = e->add_and();

    for (int j = 1; j <= r.n_inp(); ++j)
	vars.push_back(r.input_var(j));
    for (int j = 1; j <= r.n_out(); ++j)
	vars.push_back(r.output_var(j));
    for (int i = 0; i < dim-2 - in - out - nparam; ++i)
	vars.push_back(e->declare());
    for (int i = 0; i < nparam; ++i)
	vars.push_back(r.get_local(params[i]));

    for (int i = 0; i < c.size(); ++i) {
	Constraint_Handle h;
	if (c[i][0]) 
	    h = base->add_GEQ();
	else
	    h = base->add_EQ();
	for (int j = 1; j < dim-1; ++j)
	    h.update_coef(vars[j-1], c[i][j]);
	h.update_const(c[i][dim-1]);
    }
    r.finalize();
    return r;
}

Relation constraints2relation(
    int nvar, const std::vector<Global_Var_Decl *>& params,
    std::vector< std::vector<int> > & c)
{
    Relation r(nvar);
    assert(c.size() >= 1);
    int	dim = c[0].size();
    varvector vars;
    int nparam = params.size();

    F_Exists *e = r.add_exists();
    F_And *base = e->add_and();

    for (int j = 1; j <= r.n_set(); ++j)
	vars.push_back(r.set_var(j));
    for (int i = 0; i < dim-2 - nvar - nparam; ++i)
	vars.push_back(e->declare());
    for (int i = 0; i < nparam; ++i)
	vars.push_back(r.get_local(params[i]));

    for (int i = 0; i < c.size(); ++i) {
	Constraint_Handle h;
	if (c[i][0]) 
	    h = base->add_GEQ();
	else
	    h = base->add_EQ();
	for (int j = 1; j < dim-1; ++j)
	    h.update_coef(vars[j-1], c[i][j]);
	h.update_const(c[i][dim-1]);
    }
    r.finalize();
    return r;
}

Relation map2relation(pdg::UnionMap * map, 
		      const std::vector<Global_Var_Decl *>& params)
{
    assert(map->constraints.size() >= 1);
    Relation r = constraints2relation(map->input, map->output,
		    params, map->constraints[0]->el);
    for (int i = 1; i < map->constraints.size(); ++i) {
	Relation s = constraints2relation(map->input, map->output,
			params, map->constraints[i]->el);
	r = Union(r, s);
    }
    return r;
}

Relation set2relation(pdg::UnionSet * set, 
		      const std::vector<Global_Var_Decl *>& params)
{
    assert(set->constraints.size() >= 1);
    Relation r = constraints2relation(set->dim, params, set->constraints[0]->el);
    for (int i = 1; i < set->constraints.size(); ++i) {
	Relation s = constraints2relation(set->dim,
			params, set->constraints[i]->el);
	r = Union(r, s);
    }
    return r;
}

void pdg::max_index(Constraint_Handle c, varvector& vv, varvector& params)
{
    for (Constr_Vars_Iter cvi(c); cvi; ++cvi) {
	Variable_ID var = (*cvi).var;
	if (find(vv.begin(), vv.end(), var) == vv.end() &&
	    find(params.begin(), params.end(), var) == params.end())
		vv.push_back(var);
    }
}

std::vector<int> pdg::constraint_vector(Constraint_Handle c, 
				      varvector& vv, int geq)
{
    std::vector<int> v;
    v.push_back(geq);
    for (int i = 0; i < vv.size(); ++i)
	v.push_back(c.get_coef(vv[i]));
    v.push_back(c.get_const());
    return v;
}

static void fill_constraints(seq<Matrix>& constraints, Relation *s,
			     const std::vector<Global_Var_Decl *>& globals)
{
    s->simplify();

    varvector vv;
    varvector params;
    if (s->is_set())
	for (int j = 1; j <= s->n_set(); ++j)
	    vv.push_back(s->set_var(j));
    else {
	for (int j = 1; j <= s->n_inp(); ++j)
	    vv.push_back(s->input_var(j));
	for (int j = 1; j <= s->n_out(); ++j)
	    vv.push_back(s->output_var(j));
    }
    int dim = vv.size();
    for (int i = 0; i < globals.size(); ++i)
	params.push_back(s->get_local(globals[i]));

    int d = 0;
    for (DNF_Iterator di(s->query_DNF()); di; ++di, ++d) {
	vv.resize(dim);
	for (EQ_Iterator ei = (*di)->EQs(); ei; ++ei)
	    max_index((*ei), vv, params);
	for (GEQ_Iterator gi = (*di)->GEQs(); gi; ++gi)
	    max_index((*gi), vv, params);
	for (int i = 0; i < params.size(); ++i)
	    vv.push_back(params[i]);

	constraints.push_back(new pdg::Matrix());
	for (EQ_Iterator ei = (*di)->EQs(); ei; ++ei)
	    constraints[d]->el.push_back(constraint_vector((*ei), vv, 0));
	for (GEQ_Iterator gi = (*di)->GEQs(); gi; ++gi)
	    constraints[d]->el.push_back(constraint_vector((*gi), vv, 1));
    }
}

OUnionSet::OUnionSet(Relation *s, const std::vector<Global_Var_Decl *>& globals)
{
    assert (s->is_set());
    dim = s->n_set();

    fill_constraints(constraints, s, globals);
}

OUnionMap::OUnionMap(Relation *s, std::vector<Global_Var_Decl *>& globals)
{
    assert (!s->is_set());
    input = s->n_inp();
    output = s->n_out();

    fill_constraints(constraints, s, globals);
}

static at_init register_oset(OUnionSet::register_type);

void OUnionSet::register_type() {
    pdg::UnionSet::register_type();
    t_desc()[&typeid(OUnionSet)] = t_desc()[&typeid(pdg::UnionSet)];
}

static at_init register_omap(OUnionMap::register_type);

void OUnionMap::register_type() {
    pdg::UnionMap::register_type();
    t_desc()[&typeid(OUnionMap)] = t_desc()[&typeid(pdg::UnionMap)];
}

}
