#include <iostream>
#include <set>
#include "argp.h"

#include "yaml.h"
#include "pdg.h"
#include "da.h"
#include "size.h"
extern "C" {
#include "piputil.h"
}
#include "translation.h"
#include "evalue_util.h"

using pdg::PDG;
using namespace std;
using namespace da;
using namespace size;

#define ALLOC(type) (type*)malloc(sizeof(type))


#define OPT_NO_REUSE	    (BV_OPT_LAST+1)
#define OPT_NO_PROP	    (BV_OPT_LAST+2)
#define OPT_NO_SHIFT	    (BV_OPT_LAST+3)
#define OPT_NO_SCHEDULE	    (BV_OPT_LAST+4)
#define	OPT_MOVE	    (BV_OPT_LAST+5)

struct argp_option argp_options[] = {
    { "no-reuse",	    OPT_NO_REUSE,   	0,    	0 },
    { "no-propagation",	    OPT_NO_PROP,    	0,    	0 },
    { "no-shift-register",  OPT_NO_SHIFT,   	0,    	0 },
    { "no-reschedule",	    OPT_NO_SCHEDULE,	0,	0 },
    { "move",		    OPT_MOVE,		0,	0 },
#ifdef HAVE_BERNSTEIN
    { "bernstein",	    'b',		0,	0 },
#endif
    { "simulate",	    's',		0,	0 },
    { "input",		    'i',		"file",	0 },
    { "output",		    'o',		"file",	0 },
    { 0 }
};

struct arguments : size_options {
    int propagate;
    int reuse;
    int shift_register;
    int reschedule;
    int move;
    char *input;
    char *output;
};

static struct arguments arguments;

error_t parse_opt(int key, char *arg, struct argp_state *state)
{
    struct arguments *arguments = (struct arguments *)(state->input);

    switch (key) {
    case ARGP_KEY_INIT:
	state->child_inputs[0] = arguments->barvinok;
	arguments->propagate = 1;
	arguments->reuse = 1;
	arguments->shift_register = 1;
	arguments->reschedule = 1;
	arguments->move = 0;
	arguments->flags = 0;
    	arguments->input = NULL;
    	arguments->output = NULL;
	break;
    case OPT_NO_REUSE:
	arguments->reuse = 0;
	break;
    case OPT_NO_PROP:
	arguments->propagate = 0;
	break;
    case OPT_NO_SHIFT:
	arguments->shift_register = 0;
	break;
    case OPT_NO_SCHEDULE:
	arguments->reschedule = 0;
	break;
    case OPT_MOVE:
	arguments->move = 1;
	break;
    case 'b':
	arguments->flags = size_bernstein;
	break;
    case 's':
	arguments->flags = size_simulate;
	break;
    case 'i':
	arguments->input = arg;
	break;
    case 'o':
	arguments->output = arg;
	break;
    default:
	return ARGP_ERR_UNKNOWN;
    }
    return 0;
}


/* Check whether node is a propagation node 
 *
 * A propagation node simply copies a value from one array to another array.
 * In principle, a copy node could also copy a value to the same array,
 * but then we would have to be careful that the node does not recursively
 * copy values.
 * Note that the check is incomplete.  If a node performs some additional
 * operations involving only iterators, then this information is not stored
 * so we may incorrectly conclude that the node is a copy node.
 */
static bool propagation_node(pdg::node *node)
{
    pdg::statement *s = node->statement;
    if (s->accesses.size() != 2)
	return false;
    if (s->accesses[0]->type != pdg::access::read)
	return false;
    if (s->accesses[1]->type != pdg::access::write)
	return false;
    if (s->accesses[0]->array == s->accesses[1]->array)
	return false;
    if (s->accesses[1]->array->dims.size() != 0)
	return false;
    if (s->top_function)
	return false;
    return true;
}

static pdg::UnionMap * join_input(pdg::UnionMap *left, pdg::UnionMap *right)
{
    assert(left->input == right->input);
    assert(left->params.size() == right->params.size());
    int nparam = left->params.size();
    Polyhedron *D = NULL;
    for (int i = 0; i < left->constraints.size(); ++i) {
	for (int j = 0; j < right->constraints.size(); ++j) {
	    pdg::Matrix *L = left->constraints[i];
	    pdg::Matrix *R = right->constraints[j];
	    int lrow = L->el.size();
	    int rrow = R->el.size();
	    int ldim = lrow ? L->el[0].size()-2 : left->input + left->output + nparam;
	    int rdim = rrow ? R->el[0].size()-2 
			    : right->input + right->output + nparam;
	    int lextra = ldim - left->input - left->output - nparam;
	    int rextra = rdim - right->input - right->output - nparam;
	    Matrix *M = Matrix_Alloc(lrow+rrow, 2 + ldim + rdim-left->input-nparam);
	    for (int k = 0; k < lrow; ++k) {
		for (int l = 0; l < 1 + left->input + left->output; ++l)
		    value_set_si(M->p[k][l], L->el[k][l]);
		for (int l = 0; l < lextra+nparam+1; ++l)
		    value_set_si(M->p[k][1+left->input+left->output+right->output+rextra+l], 
			L->el[k][1+left->input+left->output+l]);
	    }
	    for (int k = 0; k < rrow; ++k) {
		for (int l = 0; l < 1 + right->input; ++l)
		    value_set_si(M->p[lrow+k][l], R->el[k][l]);
		for (int l = 0; l < right->output+rextra; ++l)
		    value_set_si(M->p[lrow+k][1+left->input+left->output+l], 
			R->el[k][1+right->input+l]);
		for (int l = 0; l < nparam+1; ++l)
		    value_set_si(M->p[lrow+k]
		     [1+left->input+left->output+right->output+rextra+lextra+l], 
			R->el[k][1+left->input+right->output+rextra+l]);
	    }
	    Polyhedron *P = Constraints2Polyhedron(M, POL_NO_DUAL);
	    Matrix_Free(M);
	    D = Domain_concat(Polyhedron_projectout(P, 0, left->input, nparam), D);
	    Polyhedron_Free(P);
	}
    }
    pdg::UnionMap *res = new pdg::UnionMap(left->output, right->output, D, 
					   left->params);
    Domain_Free(D);
    return res;
}

static void replace_access(pdg::statement *s, pdg::access *old_access,
			   pdg::access *new_access)
{
    s->accesses.erase(old_access);
    s->accesses.push_back(new_access);
    if (s->top_function) {
	for (int i = 0; i < s->top_function->arguments.size(); ++i)
	    if (s->top_function->arguments[i] == old_access) {
		s->top_function->arguments[i] = new_access;
		break;
	}
    }
}

static void copy_propagation(PDG* pdg)
{
    std::set<pdg::node *> to_remove_n;
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	pdg::statement *s = node->statement;
	if (!propagation_node(node))
	    continue;
	pdg::array *a = s->accesses[1]->array;
	find_deps(pdg, a, flow);
	std::set<pdg::access *> to_remove;
	for (int j = 0; j < pdg->dependences.size(); ++j) {
	    pdg::dependence *dep = pdg->dependences[j];
	    assert(dep->type == pdg::dependence::flow);
	    assert(dep->from != dep->to);
	    assert(propagation_node(dep->from));
	    if (!propagation_node(dep->from))
		continue;
	    to_remove_n.insert(dep->from);
	    pdg::UnionMap *access_map = dep->from->statement->accesses[0]->map;
	    pdg::UnionMap *prop_access = join_input(dep->relation, access_map);
	    pdg::access *access = new pdg::access;
	    access->nr = dep->to_access->nr;
	    access->array = dep->from->statement->accesses[0]->array;
	    access->type = pdg::access::read;
	    access->map = prop_access;
	    replace_access(dep->to->statement, dep->to_access, access);
	    to_remove.insert(dep->to_access);
	    dep->relation->params.resize(0);
	    dep->relation->free();
	    delete dep->relation;
	    delete dep;
	}
	std::set<pdg::access *>::iterator it;
	for (it = to_remove.begin(); it != to_remove.end(); ++it) {
	    (*it)->array = NULL;
	    (*it)->map->params.resize(0);
	    (*it)->free();
	    delete (*it);
	}
	pdg->dependences.resize(0);
    }
    std::set<pdg::node *>::iterator itn;
    for (itn = to_remove_n.begin(); itn != to_remove_n.end(); ++itn) {
	(*itn)->statement->accesses[0]->array = NULL;
	(*itn)->statement->accesses[0]->map->params.resize(0);
	(*itn)->statement->accesses[1]->array = NULL;
	(*itn)->statement->accesses[1]->map->params.resize(0);
	(*itn)->source->params.resize(0);
	(*itn)->free();
	delete (*itn);
	pdg->nodes.erase(*itn);
    }
}

/* Check whether dep writes and reads from consecutive iterations of
 * the same domain.
 * We check this property by testing whether any other read for the
 * same access occurs in between.
 */
bool is_consecutive_dep(PDG *pdg, pdg::dependence *dep, Polyhedron *P, 
			Polyhedron *ctx)
{
    assert(dep->to == dep->from);
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *other_dep = pdg->dependences[i];
	if (other_dep == dep)
	    continue;
	if (other_dep->to != dep->to)
	    continue;
	if (other_dep->to_access != dep->to_access)
	    continue;
	pdg::UnionMap *other_rel = other_dep->relation;
	Polyhedron *other_P = *other_rel;
	bool consec = pip_is_consecutive_dep(P, other_P, other_rel->input,
					     other_rel->output, ctx);
	Domain_Free(other_P);
	if (!consec)
	    return false;
    }
    return true;
}

/* Computes the set of iterations where the fifo is both read and written to */
static Polyhedron *fifo_propagation_region(pdg::UnionMap *rel, Polyhedron *dep)
{
    int nparam = rel->params.size();
    Polyhedron *reads = NULL;
    Polyhedron *inter_tmp = NULL, *inter;
    assert(rel->input == rel->output);
    for (Polyhedron *P = dep; P; P = P->next) {
	Polyhedron *read = Polyhedron_projectout_dep(P, rel->input);
	reads = Domain_concat(read, reads);
    }
    for (Polyhedron *P = dep; P; P = P->next) {
	for (Polyhedron *read = reads; read; read = read->next) {
	    Polyhedron *I = EPoly_Intersection(P, read, rel->input, nparam);
	    inter_tmp = Domain_concat(I, inter_tmp);
	}
    }
    Domain_Free(reads);
    inter = Domain_projectout(inter_tmp, rel->input, rel->output, nparam);
    Domain_Free(inter_tmp);
    return inter;
}
/* An auxiliary class that keeps a reference to a Polyhedron
 * and frees it when it goes out of scope.
 */
struct temp_Domain {
    Polyhedron * const D;
    temp_Domain(Polyhedron *P) : D(P) {}
    operator Polyhedron* () const { return D; }
    ~temp_Domain() {
	Domain_Free(D);
    }
};

/* Check if we can use a shift register for the self-dependence dep
 * on the domain D.
 * If yes, return the required size of the shift register.
 * If no, return NULL.
 *
 * We assume that the register is shifted on every iteration of D.
 * If the number of shifts between any pair of writes and reads is
 * constant, then we can use a shift register.
 * Note that the size of the shift register may be larger than
 * that of the FIFO, since there may be iterations of the domain
 * between corresponding writes and reads where nothing is write
 * to or read from the shift register.
 */
static size::enumerator *shift_register_size(pdg::PDG *pdg, Polyhedron *D,
					     Polyhedron *dep,
					     int dim, Polyhedron *ctx)
{
    int nparam = ctx->Dimension;
    evalue *old = NULL;
    assert(!ctx->next);
    for (Polyhedron *R = dep; R; R = R->next) {
	Polyhedron *read = Polyhedron_projectout_dep(R, dim);
	Polyhedron *read_min;
	read_min = pip_lexminmax(read, 0, read->Dimension-nparam, nparam, 0);
	evalue *res = ALLOC(evalue);
	bool shift = true;

	value_init(res->d);
	value_set_si(res->d, 0);
	res->x.p = new_enode(::partition, 2, read->Dimension);
	EVALUE_SET_DOMAIN(res->x.p->arr[0], read);
	evalue_set_si(&res->x.p->arr[1], 1, 1);

	for (Polyhedron *P = D; P; P = P->next)
	    res = evalue_lexbetween_poly(P, R, dim, ctx, POL_NO_DUAL, res);

	reduce_evalue(res);

	if (res->x.p->size/2 > 1)
	    shift = false;
	else if (!evalue_is_constant(res, read_min->Dimension - nparam))
	    shift = false;
	if (shift) {
	    evalue_frac2floor(res);
	    evalue_remove_iterators(res, read_min, read_min->Dimension - nparam, ctx);
	    if (old) {
	    	if (!eequal(&res->x.p->arr[1], &old->x.p->arr[1]))
		    shift = false;
		else {
		    Polyhedron *old_domain = EVALUE_DOMAIN(old->x.p->arr[0]);
		    Polyhedron *res_domain = EVALUE_DOMAIN(res->x.p->arr[0]);
		    Polyhedron *domain = DomainUnion(old_domain, res_domain, 
						     POL_NO_DUAL);
		    EVALUE_SET_DOMAIN(old->x.p->arr[0], domain);
		    Domain_Free(old_domain);
		}
	    }
	}
	Polyhedron_Free(read_min);
	if (!shift) {
	    free_evalue_refs(res);
	    free(res);
	    if (old) {
		free_evalue_refs(old);
		free(old);
	    }
	    return NULL;
	}
	if (!old)
	    old = res;
	else {
	    free_evalue_refs(res);
	    free(res);
	}
    }
    assert(old);
    return new size::enumerator(old, pdg);
}

int main(int argc, char * argv[])
{
    FILE *in = stdin, *out = stdout;
    int c, ind = 0;
    PDG *pdg;
    bool output_allocated = false;
    static struct argp_child argp_children[] = {
	{ &barvinok_argp,    	0,	"barvinok options",	1 },
	{ 0 }
    };
    static struct argp argp = { argp_options, parse_opt, 0, 0, argp_children };
    struct barvinok_options *options = barvinok_options_new_with_defaults();

    arguments.barvinok = options;
    argp_parse(&argp, argc, argv, 0, 0, &arguments);

    if (!arguments.reuse)
	arguments.propagate = 0;

    if (arguments.input && strcmp(arguments.input, "-")) {
	in = fopen(arguments.input, "r");
	assert(in);
	if (!arguments.output) {
	    int len = strlen(arguments.input);
	    if (len > 5 && !strcmp(arguments.input+len-5, ".yaml"))
		len -= 5;
	    arguments.output = new char[len+9+1];
	    output_allocated = true;
	    strncpy(arguments.output, arguments.input, len);
	    strcpy(arguments.output+len, arguments.reuse ? "_pn.yaml" : "_cpn.yaml");
	}
    }

    pdg = yaml::Load<PDG>(in);
    int nparam = pdg->params.size();

    if (arguments.propagate)
	copy_propagation(pdg);

    da::type da_type = pn_propagate;
    if (!arguments.propagate)
	da_type = pn;
    if (!arguments.reuse)
	da_type = flow;
    for (int i = 0; i < pdg->arrays.size(); ++i)
	find_deps(pdg, pdg->arrays[i], da_type);

    bool need_translation = false;
    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);
    if (pdg->context) {
	Polyhedron *min_values = NULL;
	min_values = pip_lexminmax(ctx, 0, nparam, 0, 0);
	for (int i = 0; i < min_values->NbEq; ++i) {
	    int pos = First_Non_Zero(min_values->Constraint[i]+1, nparam);
	    assert(value_one_p(min_values->Constraint[i][1+pos]));
	    if (pdg->params[pos]->value)
		continue;
	    pdg->params[pos]->value = 
		new integer(-VALUE_TO_INT(min_values->Constraint[i][1+nparam]));
	}
	Polyhedron_Free(min_values);
    }
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized) {
	    fprintf(stderr, "access to uninitialized data from %s on line %d\n",
			     dep->array->name->s.c_str(), 
			     dep->to->statement->line);
	    dep->relation->Dump(stderr);
	    continue;
	}
	pdg::UnionMap *rel = dep->relation;
	temp_Domain P(*rel);
	if (dep->type == pdg::dependence::pn_part ||
	    dep->type == pdg::dependence::pn_wire) {
	    dep->reordering = 0;
	    dep->multiplicity = 0;
	    dep->size = new pdg::enumerator(0);
	    dep->value_size = new integer(0);
	    continue;
	}
	dep->reordering = pip_is_reordering(P, rel->input, rel->output, ctx);
	dep->multiplicity = 0;
	if (!arguments.reuse)
	    dep->multiplicity = pip_has_multiplicity(P, rel->input, 
						     rel->output, ctx);
	if (dep->reordering || dep->multiplicity)
	    continue;
	if (dep->to != dep->from) {
	    need_translation = true;
	    continue;
	}
	if (pip_is_consecutive_loop(P, rel->input, rel->output, ctx)) {
	    dep->size = new pdg::enumerator(1);
	    dep->value_size = new integer(1);
	    if (dep->from_access == dep->to_access &&
		is_consecutive_dep(pdg, dep, P, ctx))
		dep->type = pdg::dependence::pn_hold;
	} else {
	    size::enumerator *size = NULL;
	    temp_Domain D(*dep->to->source);
	    if (arguments.shift_register)
		size = shift_register_size(pdg, D, P, dep->to->source->dim, ctx);
	    if (size)
		dep->type = pdg::dependence::pn_shift;
	    else
		size = selfloop_size(pdg, P, rel->input, rel->output, 
				     ctx, &arguments);
	    if (size) {
		dep->size = *size;
		if (pdg->context)
		    dep->value_size = size->evaluate();
		delete size;
	    }
	}
    }

    bool translated = need_translation;
    if (arguments.reschedule) {
	common_dimension(pdg);
	if (need_translation)
	    translated = translate(pdg, arguments.move);
    } else {
	set_scattering_from_prefix(pdg);
	translated = true;
    }
    if (translated) {
	for (int i = 0; i < pdg->dependences.size(); ++i) {
	    pdg::dependence *dep = pdg->dependences[i];
	    if (dep->type == pdg::dependence::uninitialized)
		continue;
	    if (dep->reordering)
		continue;
	    if (dep->multiplicity)
		continue;
	    if (dep->to == dep->from)
		continue;
	    if (dep->size)
		continue;
	    pdg::UnionMap *rel = scatter_dependence(dep);
	    Polyhedron *P = *rel;
	    if (pip_is_consecutive_loop(P, rel->input, rel->output, ctx)) {
		dep->size = new pdg::enumerator(1);
		dep->value_size = new integer(1);
	    } else {
		size::enumerator *size;
		size = selfloop_size(pdg, P, rel->input, rel->output, ctx,
				     &arguments);
		if (size) {
		    dep->size = *size;
		    if (pdg->context)
			dep->value_size = size->evaluate();
		    delete size;
		}
	    }
	    Domain_Free(P);
	    rel->free();
	    delete rel;
	}
    }
    Polyhedron_Free(ctx);

    pdg->add_history_line("pn", argc, argv);

    if (arguments.output && strcmp(arguments.output, "-")) {
	out = fopen(arguments.output, "w");
	assert(out);
    }
    if (output_allocated)
	delete [] arguments.output;

    pdg->Dump(out);
    pdg->free();
    delete pdg;
    barvinok_options_free(options);

    return 0;
}
